/**
 * @package koa-files
 * @license MIT
 * @version 2.3.1
 * @author nuintun <nuintun@qq.com>
 * @description A static files serving middleware for koa.
 * @see https://github.com/nuintun/koa-files#readme
 */

import etag from 'etag';
import destroy from 'destroy';
import { PassThrough } from 'stream';
import fs from 'fs';
import { resolve, join, extname } from 'path';
import parseRange from 'range-parser';
import { decodeURI, unixify, isETagFresh, isETag, boundaryGenerator, isOutRoot, fstat, hasTrailingSlash } from './utils.js';

/**
 * @module Send
 * @license MIT
 * @author nuintun
 */
/**
 * @class Send
 */
class Send {
  /**
   * @constructor
   * @param {Context} ctx
   * @param {string} root
   * @param {Options} options
   */
  constructor(ctx, root = '.', options) {
    let { cacheControl } = options;
    const path = decodeURI(ctx.path);
    this.ctx = ctx;
    this.root = unixify(resolve(root));
    this.options = Object.assign(Object.assign({}, options), { cacheControl });
    this.path = path === -1 ? -1 : unixify(join(this.root, path));
  }
  /**
   * @method isConditionalGET
   * @returns {boolean}
   */
  isConditionalGET() {
    const { request } = this.ctx;
    return !!(
      request.get('If-Match') ||
      request.get('If-None-Match') ||
      request.get('If-Modified-Since') ||
      request.get('if-Unmodified-Since')
    );
  }
  /**
   * @method isPreconditionFailure
   * @returns {boolean}
   */
  isPreconditionFailure() {
    const { request, response } = this.ctx;
    // If-Match
    const match = request.get('If-Match');
    if (match) {
      const etag = response.get('ETag');
      return !etag || (match !== '*' && !isETagFresh(match, etag));
    }
    // If-Unmodified-Since
    const unmodifiedSince = Date.parse(request.get('If-Unmodified-Since'));
    if (!isNaN(unmodifiedSince)) {
      const lastModified = Date.parse(response.get('Last-Modified'));
      return isNaN(lastModified) || lastModified > unmodifiedSince;
    }
    return false;
  }
  /**
   * @method isRangeFresh
   * @returns {boolean}
   */
  isRangeFresh() {
    const { request, response } = this.ctx;
    const ifRange = request.get('If-Range');
    // No If-Range
    if (!ifRange) {
      return true;
    }
    // If-Range as etag
    if (isETag(ifRange)) {
      const etag = response.get('ETag');
      return !!(etag && isETagFresh(ifRange, etag));
    }
    // If-Range as modified date
    const lastModified = response.get('Last-Modified');
    return Date.parse(lastModified) <= Date.parse(ifRange);
  }
  /**
   * @method isIgnore
   * @param {string} path
   * @returns {boolean}
   */
  isIgnore(path) {
    const { ignore } = this.options;
    return (typeof ignore === 'function' ? ignore(path) : false) === true;
  }
  /**
   * @method parseRange
   * @param {Stats} stats
   * @returns {Ranges}
   */
  parseRange(stats) {
    const { ctx } = this;
    const result = [];
    const { size } = stats;
    const { request } = ctx;
    // Content-Length
    let contentLength = size;
    // Range support
    if (this.options.acceptRanges !== false) {
      const range = request.get('Range');
      // Range fresh
      if (range && this.isRangeFresh()) {
        // Parse range -1 -2 or []
        const ranges = parseRange(size, range, { combine: true });
        // -1 signals an unsatisfiable range
        // -2 signals a malformed header string
        if (ranges === -1 || ranges === -2) {
          return ranges;
        }
        // Ranges ok, support multiple ranges
        if (ranges.type === 'bytes') {
          // Set 206 status
          ctx.status = 206;
          // Multiple ranges
          if (ranges.length > 1) {
            // Reset content-length
            contentLength = 0;
            // Range boundary
            const boundary = `<${boundaryGenerator()}>`;
            const suffix = `\r\n--${boundary}--\r\n`;
            const contentType = `Content-Type: ${ctx.type}`;
            ctx.type = `multipart/byteranges; boundary=${boundary}`;
            // Map ranges
            ranges.forEach(({ start, end }, index) => {
              // The first prefix boundary no \r\n
              const prefixHead = index > 0 ? '\r\n' : '';
              const contentRange = `Content-Range: bytes ${start}-${end}/${size}`;
              const prefix = `${prefixHead}--${boundary}\r\n${contentType}\r\n${contentRange}\r\n\r\n`;
              // Compute content-length
              contentLength += end - start + 1 + Buffer.byteLength(prefix);
              // Cache range
              result.push({ start, end, prefix });
            });
            // The last add suffix boundary
            result[result.length - 1].suffix = suffix;
            // Compute content-length
            contentLength += Buffer.byteLength(suffix);
          } else {
            const { start, end } = ranges[0];
            ctx.set('Content-Range', `bytes ${start}-${end}/${size}`);
            // Compute content-length
            contentLength = end - start + 1;
            // Cache range
            result.push({ start, end });
          }
        }
      }
    }
    // Set Content-Length
    ctx.length = contentLength;
    return result.length ? result : [{ start: 0 }];
  }
  /**
   * @method setupHeaders
   * @param {string} path
   * @param {Stats} stats
   */
  setupHeaders(path, stats) {
    const { ctx, options } = this;
    const { toString } = Object.prototype;
    const { acceptRanges, cacheControl, lastModified } = options;
    // Set status
    ctx.status = 200;
    // Set Content-Type
    ctx.type = extname(path);
    // ETag
    if (options.etag !== false) {
      // Set ETag
      ctx.set('ETag', etag(stats));
    }
    // Accept-Ranges
    if (acceptRanges !== false) {
      // Set Accept-Ranges
      ctx.set('Accept-Ranges', 'bytes');
    }
    // Last-Modified
    if (lastModified !== false) {
      // Set mtime utc string
      ctx.set('Last-Modified', stats.mtime.toUTCString());
    }
    // Cache-Control
    if (cacheControl && toString.call(cacheControl) === '[object String]') {
      // Set Cache-Control
      ctx.set('Cache-Control', cacheControl);
    }
  }
  /**
   * @method read
   * @param {string} path
   * @param {Range} range
   * @param {PassThrough} buffer
   * @param {boolean} end
   * @returns {Promise<true>}
   */
  read(path, range, buffer, end) {
    return new Promise((resolve, reject) => {
      // Create file stream
      const file = fs.createReadStream(path, range);
      // File read stream open
      if (range.prefix) {
        file.once('open', () => {
          // Write prefix boundary
          buffer.write(range.prefix);
        });
      }
      // File read stream error
      file.once('error', error => {
        // Unpipe
        file.unpipe(buffer);
        // Destroy file stream
        destroy(file);
        // Reject
        reject(error);
      });
      // File read stream end
      if (range.suffix) {
        file.once('end', () => {
          // Push suffix boundary
          buffer.write(range.suffix);
        });
      }
      // File read stream close
      file.once('close', () => {
        // Unpipe
        file.unpipe(buffer);
        // Destroy file stream
        destroy(file);
        // Resolve
        resolve(true);
      });
      // Write data to buffer
      file.pipe(buffer, { end });
    });
  }
  /**
   * @method send
   * @param {string} path
   * @param {Range[]} ranges
   */
  async send(path, ranges) {
    // Ranges count
    let count = ranges.length;
    // Set stream body, highWaterMark 64kb
    const stream = new PassThrough({ highWaterMark: 65536 });
    // Set response body
    this.ctx.body = stream;
    // Read file ranges
    try {
      for (const range of ranges) {
        await this.read(path, range, stream, --count === 0);
      }
    } catch (error) {
      // End stream when read exception
      stream.end();
    }
  }
  /**
   * @method response
   * @returns {Promise<boolean>}
   */
  async response() {
    const { ctx, root, path } = this;
    // Only support GET and HEAD (405)
    if (ctx.method !== 'GET' && ctx.method !== 'HEAD') {
      return false;
    }
    // Path -1 or null byte(s)
    if (path === -1 || path.includes('\0')) {
      return ctx.throw(400);
    }
    // Malicious path (403)
    if (isOutRoot(path, root)) {
      return false;
    }
    // Is ignore path or file (403)
    if (this.isIgnore(path)) {
      return false;
    }
    // File stats
    let stats;
    // Get file stats
    try {
      stats = await fstat(path);
    } catch (error) {
      // 404 | 500
      return false;
    }
    // File not exist (404 | 500)
    if (!stats) {
      return false;
    }
    // Is directory (403)
    if (stats.isDirectory()) {
      return false;
    }
    // Not a directory but has trailing slash (404)
    if (hasTrailingSlash(path)) {
      return false;
    }
    // Setup headers
    this.setupHeaders(path, stats);
    // Conditional get support
    if (this.isConditionalGET()) {
      // Request precondition failure
      if (this.isPreconditionFailure()) {
        return ctx.throw(412);
      }
      // Request fresh (304)
      if (ctx.fresh) {
        // Set status
        ctx.status = 304;
        // Set body null
        ctx.body = null;
        return true;
      }
    }
    // Head request
    if (ctx.method === 'HEAD') {
      // Set content-length
      ctx.length = stats.size;
      // Set body null
      ctx.body = null;
      return true;
    }
    // Parse ranges
    const ranges = this.parseRange(stats);
    // 416
    if (ranges === -1) {
      // Set content-range
      ctx.set('Content-Range', `bytes */${stats.size}`);
      // Unsatisfiable 416
      return ctx.throw(416);
    }
    // 400
    if (ranges === -2) {
      return ctx.throw(400);
    }
    // Send file
    this.send(path, ranges);
    return true;
  }
}

export { Send as default };
