/**
 * @package koa-files
 * @license MIT
 * @version 2.3.1
 * @author nuintun <nuintun@qq.com>
 * @description A static files serving middleware for koa.
 * @see https://github.com/nuintun/koa-files#readme
 */

import fs from 'fs';
import { relative, isAbsolute } from 'path';

/**
 * @module utils
 * @license MIT
 * @author nuintun
 */
const CHARS = Array.from('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz');
/**
 * @function isOutRoot
 * @description Test path is out of bound of root
 * @param {string} path
 * @param {string} root
 * @returns {boolean}
 */
function isOutRoot(path, root) {
  path = relative(root, path);
  return /\.\.(?:[\\/]|$)/.test(path) || isAbsolute(path);
}
/**
 * @function unixify
 * @description Convert path separators to posix/unix-style forward slashes
 * @param {string} path
 * @returns {string}
 */
function unixify(path) {
  return path.replace(/\\/g, '/');
}
/**
 * @function decodeURI
 * @description Decode URI component.
 * @param {string} URI
 * @returns {string|-1}
 */
function decodeURI(URI) {
  try {
    return decodeURIComponent(URI);
  } catch (error) {
    return -1;
  }
}
/**
 * @function boundaryGenerator
 * @description Create boundary
 * @returns {string}
 */
function boundaryGenerator() {
  let boundary = '';
  // Create boundary
  for (let i = 0; i < 38; i++) {
    boundary += CHARS[Math.floor(Math.random() * 62)];
  }
  // Return boundary
  return boundary;
}
/**
 * @function fstat
 * @param {string} path
 * @returns {Promise<Stats>}
 */
function fstat(path) {
  return new Promise((resolve, reject) => {
    fs.stat(path, (error, stats) => {
      error ? reject(error) : resolve(stats);
    });
  });
}
/**
 * @function hasTrailingSlash
 * @param {string} path
 * @returns {boolean}
 */
function hasTrailingSlash(path) {
  return /\/$/.test(path);
}
/**
 * @function parseTokens
 * @description Parse a HTTP tokens.
 * @param {string[]} value
 */
function parseTokens(value) {
  let start = 0;
  let end = 0;
  let tokens = [];
  // gather tokens
  for (let i = 0, length = value.length; i < length; i++) {
    switch (value.charCodeAt(i)) {
      case 0x20:
        // ' '
        if (start === end) {
          start = end = i + 1;
        }
        break;
      case 0x2c:
        // ','
        tokens.push(value.substring(start, end));
        start = end = i + 1;
        break;
      default:
        end = i + 1;
        break;
    }
  }
  // final token
  tokens.push(value.substring(start, end));
  return tokens;
}
/**
 * @function isETagFresh
 * @param {string} match
 * @param {string} etag
 * @returns {boolean}
 */
function isETagFresh(match, etag) {
  return parseTokens(match).some(match => {
    return match === etag || match === 'W/' + etag || 'W/' + match === etag;
  });
}
/**
 * @function isETag
 * @param {string} etag
 * @returns {boolean}
 */
function isETag(etag) {
  return /(?:W\/)?"(?:[ !#-\x7E\x80-\xFF]*|\r\n[\t ]|\\.)*"/.test(etag);
}

export { boundaryGenerator, decodeURI, fstat, hasTrailingSlash, isETag, isETagFresh, isOutRoot, unixify };
